@@ -1,6 +1,5 @@
 import logging
 import os
-import shlex
 import sys
 import warnings
 from collections.abc import Callable, Sequence
@@ -14,6 +13,7 @@ from configparser import (
 from functools import wraps
 from io import StringIO
 from pathlib import Path
+from shlex import shlex
 
 from pytensor.utils import hash_from_code
 
@@ -81,6 +81,8 @@ class PyTensorConfigParser:
     allow_gc: bool
     optimizer: str
     optimizer_verbose: bool
+    optimizer_verbose_ignore: str
+    compiler_verbose: bool
     on_opt_error: str
     nocleanup: bool
     on_unused_input: str
@@ -155,7 +157,6 @@ class PyTensorConfigParser:
     scan__allow_gc: bool
     scan__allow_output_prealloc: bool
     # add_numba_configvars
-    numba__vectorize_target: str
     numba__fastmath: bool
     numba__cache: bool
     # add_caching_dir_configvars
@@ -409,7 +410,9 @@ class ConfigParam:
                 f"The config parameter '{self.name}' was registered on a different instance of the PyTensorConfigParser."
                 f" It is not accessible through the instance with id '{id(cls)}' because of safeguarding."
             )
-        if not hasattr(self, "val"):
+        try:
+            return self.val
+        except AttributeError:
             try:
                 val_str = cls.fetch_val_for_key(self.name, delete_key=delete_key)
                 self.is_default = False
@@ -541,7 +544,7 @@ def parse_config_string(
     Parses a config string (comma-separated key=value components) into a dict.
     """
     config_dict = {}
-    my_splitter = shlex.shlex(config_string, posix=True)
+    my_splitter = shlex(config_string, posix=True)
     my_splitter.whitespace = ","
     my_splitter.whitespace_split = True
     for kv_pair in my_splitter:
