@@ -77,12 +77,15 @@ Optimizations associated with these BLAS Ops are in tensor.rewriting.blas
 
 import functools
 import logging
-import os
 import shlex
-import time
+import warnings
 from pathlib import Path
 
 import numpy as np
+from numpy.lib.array_utils import normalize_axis_tuple
+from scipy.linalg import get_blas_funcs
+
+from pytensor.graph import Variable, vectorize_graph
 
 
 try:
@@ -93,73 +96,58 @@ except ImportError:
 
 import pytensor.scalar
 from pytensor.configdefaults import config
-from pytensor.graph.basic import Apply, view_roots
+from pytensor.graph.basic import Apply
 from pytensor.graph.op import Op
 from pytensor.graph.utils import InconsistencyError, MethodNotDefined, TestValueError
 from pytensor.link.c.op import COp
 from pytensor.link.c.params_type import ParamsType
 from pytensor.printing import FunctionPrinter, pprint
 from pytensor.scalar import bool as bool_t
-from pytensor.tensor import basic as ptb
-from pytensor.tensor.basic import expand_dims
+from pytensor.tensor.basic import as_tensor_variable, cast
 from pytensor.tensor.blas_headers import blas_header_text, blas_header_version
-from pytensor.tensor.elemwise import DimShuffle
-from pytensor.tensor.math import add, mul, neg, sub, variadic_add
-from pytensor.tensor.shape import shape_padright, specify_broadcastable
-from pytensor.tensor.type import DenseTensorType, TensorType, integer_dtypes, tensor
+from pytensor.tensor.math import dot, tensordot
+from pytensor.tensor.shape import specify_broadcastable
+from pytensor.tensor.type import DenseTensorType, tensor
 
 
 _logger = logging.getLogger("pytensor.tensor.blas")
 
-try:
-    import scipy.linalg.blas
 
-    have_fblas = True
-    try:
-        fblas = scipy.linalg.blas.fblas
-    except AttributeError:
-        # A change merged in Scipy development version on 2012-12-02 replaced
-        # `scipy.linalg.blas.fblas` with `scipy.linalg.blas`.
-        # See http://github.com/scipy/scipy/pull/358
-        fblas = scipy.linalg.blas
-    _blas_gemv_fns = {
-        np.dtype("float32"): fblas.sgemv,
-        np.dtype("float64"): fblas.dgemv,
-        np.dtype("complex64"): fblas.cgemv,
-        np.dtype("complex128"): fblas.zgemv,
-    }
-except ImportError as e:
-    have_fblas = False
-    # This is used in Gemv and ScipyGer. We use CGemv and CGer
-    # when config.blas__ldflags is defined. So we don't need a
-    # warning in that case.
-    if not config.blas__ldflags:
-        _logger.warning(
-            "Failed to import scipy.linalg.blas, and "
-            "PyTensor flag blas__ldflags is empty. "
-            "Falling back on slower implementations for "
-            "dot(matrix, vector), dot(vector, matrix) and "
-            f"dot(vector, vector) ({e!s})"
-        )
-
-
-# If check_init_y() == True we need to initialize y when beta == 0.
-def check_init_y():
-    if check_init_y._result is None:
-        if not have_fblas:  # pragma: no cover
-            check_init_y._result = False
+def view_roots(node: Variable) -> list[Variable]:
+    """Return the leaves from a search through consecutive view-maps."""
+    owner = node.owner
+    if owner is not None:
+        try:
+            vars_to_views = {owner.outputs[o]: i for o, i in owner.op.view_map.items()}
+        except AttributeError:
+            return [node]
+        if node in vars_to_views:
+            answer = []
+            for i in vars_to_views[node]:
+                answer += view_roots(owner.inputs[i])
+            return answer
         else:
-            y = float("NaN") * np.ones((2,))
-            x = np.ones((2,))
-            A = np.ones((2, 2))
-            gemv = _blas_gemv_fns[y.dtype]
-            gemv(1.0, A.T, x, 0.0, y, overwrite_y=True, trans=True)
-            check_init_y._result = np.isnan(y).any()
-
-    return check_init_y._result
+            return [node]
+    else:
+        return [node]
 
 
-check_init_y._result = None  # type: ignore
+def must_initialize_y_gemv():
+    # Check whether Scipy GEMV could output nan if y in not initialized
+    from scipy.linalg.blas import get_blas_funcs
+
+    if must_initialize_y_gemv._result is None:
+        y = np.full((2,), np.nan)
+        x = np.ones((2,))
+        A = np.ones((2, 2))
+        gemv = get_blas_funcs("gemv", dtype=y.dtype)
+        gemv(1.0, A.T, x, 0.0, y, overwrite_y=True, trans=True)
+        must_initialize_y_gemv._result = np.isnan(y).any()
+
+    return must_initialize_y_gemv._result
+
+
+must_initialize_y_gemv._result = None  # type: ignore
 
 
 class Gemv(Op):
@@ -187,11 +175,11 @@ class Gemv(Op):
             return f"{self.__class__.__name__}{{no_inplace}}"
 
     def make_node(self, y, alpha, A, x, beta):
-        y = ptb.as_tensor_variable(y)
-        x = ptb.as_tensor_variable(x)
-        A = ptb.as_tensor_variable(A)
-        alpha = ptb.as_tensor_variable(alpha)
-        beta = ptb.as_tensor_variable(beta)
+        y = as_tensor_variable(y)
+        x = as_tensor_variable(x)
+        A = as_tensor_variable(A)
+        alpha = as_tensor_variable(alpha)
+        beta = as_tensor_variable(beta)
         if y.dtype != A.dtype or y.dtype != x.dtype:
             raise TypeError(
                 "Gemv requires matching dtypes", (y.dtype, A.dtype, x.dtype)
@@ -211,14 +199,15 @@ class Gemv(Op):
         return Apply(self, inputs, [y.type()])
 
     def perform(self, node, inputs, out_storage):
+        from scipy.linalg.blas import get_blas_funcs
+
         y, alpha, A, x, beta = inputs
         if (
-            have_fblas
-            and y.shape[0] != 0
+            y.shape[0] != 0
             and x.shape[0] != 0
-            and y.dtype in _blas_gemv_fns
+            and y.dtype in {"float32", "float64", "complex64", "complex128"}
         ):
-            gemv = _blas_gemv_fns[y.dtype]
+            gemv = get_blas_funcs("gemv", dtype=y.dtype)
 
             if A.shape[0] != y.shape[0] or A.shape[1] != x.shape[0]:
                 raise ValueError(
@@ -226,7 +215,13 @@ class Gemv(Op):
                     f"(beta * y + alpha * dot(A, x)). y: {y.shape}, A: {A.shape}, x: {x.shape}"
                 )
 
-            if beta == 0 and check_init_y():
+            if beta == 0 and must_initialize_y_gemv():
+                # Most BLAS implementations of GEMV ignore y=nan when beta=0
+                # PyTensor considers that the correct behavior,
+                # and even exploits it to avoid copying or initializing outputs.
+                # By deciding to exploit this, however, it becomes our responsibility
+                # to ensure the behavior even in the rare cases BLAS deviates,
+                # or users will get errors, even for graphs that had no nan to begin with.
                 y.fill(0)
 
             # Here I suppose that A is in c order. If we don't make it
@@ -286,10 +281,10 @@ class Ger(Op):
             return f"{self.__class__.__name__}{{non-destructive}}"
 
     def make_node(self, A, alpha, x, y):
-        A = ptb.as_tensor_variable(A)
-        y = ptb.as_tensor_variable(y)
-        x = ptb.as_tensor_variable(x)
-        alpha = ptb.as_tensor_variable(alpha)
+        A = as_tensor_variable(A)
+        y = as_tensor_variable(y)
+        x = as_tensor_variable(x)
+        alpha = as_tensor_variable(alpha)
         if not (A.dtype == x.dtype == y.dtype == alpha.dtype):
             raise TypeError(
                 "ger requires matching dtypes", (A.dtype, alpha.dtype, x.dtype, y.dtype)
@@ -312,18 +307,17 @@ class Ger(Op):
 
         return Apply(self, inputs, [A.type()])
 
-    def perform(self, node, inp, out):
-        cA, calpha, cx, cy = inp
-        (cZ,) = out
-        if self.destructive:
-            A = cA
-        else:
-            A = cA.copy()
-        if calpha != 1:
-            A += calpha * np.outer(cx, cy)
-        else:
-            A += np.outer(cx, cy)
-        cZ[0] = A
+    def perform(self, node, inputs, output_storage):
+        A, alpha, x, y = inputs
+        if A.size:
+            # GER doesn't handle zero-sized inputs
+            ger_func = get_blas_funcs("ger", dtype=A.dtype)
+            if A.flags["C_CONTIGUOUS"]:
+                # Work on transposed system to avoid copying
+                A = ger_func(alpha, y, x, a=A.T, overwrite_a=self.destructive).T
+            else:
+                A = ger_func(alpha, x, y, a=A, overwrite_a=self.destructive)
+        output_storage[0][0] = A
 
     def infer_shape(self, fgraph, node, input_shapes):
         return [input_shapes[0]]
@@ -407,9 +401,9 @@ def _ldflags(
             include_dir=False,
         )
         for d in dirs:
-            for f in os.listdir(d.strip('"')):
-                if f.endswith(".so") or f.endswith(".dylib") or f.endswith(".dll"):
-                    if any(f.find(ll) >= 0 for ll in l):
+            for f in Path(d.strip('"')).iterdir():
+                if f.suffix in {".so", ".dylib", ".dll"}:
+                    if any(f.stem.find(ll) >= 0 for ll in l):
                         found_dyn = True
         # Special treatment of clang framework. Specifically for MacOS Accelerate
         if "-framework" in l and "Accelerate" in l:
@@ -489,13 +483,6 @@ class GemmRelated(COp):
         #ifndef MOD
         #define MOD %
         #endif
-        static double time_time() // a time function like time.perf_counter()
-        {
-            struct timeval tv;
-            gettimeofday(&tv, 0);
-            return (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;
-        }
-
         void compute_strides(npy_intp *shape, int N_shape, int type_size, npy_intp *res) {
             int s;
             res[N_shape - 1] = type_size;
@@ -508,9 +495,7 @@ class GemmRelated(COp):
         return blas_header_text() + mod_str
 
     def c_headers(self, **kwargs):
-        # std.cout doesn't require the '%' symbol to print stuff...
-        # so it works much better with python's string-substitution stuff.
-        return ["<iostream>", "<time.h>", "<sys/time.h>"]
+        return []
 
     def c_libraries(self, **kwargs):
         return ldflags()
@@ -531,7 +516,7 @@ class GemmRelated(COp):
         int unit = 0;
 
         int type_num = PyArray_DESCR(%(_x)s)->type_num;
-        int type_size = PyArray_DESCR(%(_x)s)->elsize; // in bytes
+        int type_size = PyArray_ITEMSIZE(%(_x)s); // in bytes
 
         npy_intp* Nx = PyArray_DIMS(%(_x)s);
         npy_intp* Ny = PyArray_DIMS(%(_y)s);
@@ -719,8 +704,6 @@ class GemmRelated(COp):
                 char N = 'N';
                 char T = 'T';
                 int Nz0 = Nz[0], Nz1 = Nz[1], Nx1 = Nx[1];
-                //std::cerr << (unit/256) MOD 16 << (unit / 16) MOD 16 << unit MOD 16<< '\\n';
-                //double t0 = time_time();
                 switch(unit)
                 {
                     case 0x000: sgemm_(&N, &N, &Nz1, &Nz0, &Nx1, &a, y, &sy_0, x, &sx_0, &b, z, &sz_0); break;
@@ -733,7 +716,6 @@ class GemmRelated(COp):
                     case 0x111: sgemm_(&N, &N, &Nz0, &Nz1, &Nx1, &a, x, &sx_1, y, &sy_1, &b, z, &sz_1); break;
                     default: PyErr_SetString(PyExc_ValueError, "some matrix has no unit stride"); %(fail)s;
                 };
-                //fprintf(stderr, "Calling sgemm %%i %%i %%i %%i took %%f\\n", unit, Nz1, Nz0, Nx1, time_time() - t0);
         """
 
     case_double = """
@@ -752,14 +734,6 @@ class GemmRelated(COp):
                 char N = 'N';
                 char T = 'T';
                 int Nz0 = Nz[0], Nz1 = Nz[1], Nx1 = Nx[1];
-                //std::cerr << (unit/256) MOD 16 << (unit / 16) MOD 16 << unit MOD 16<< '\\n';
-                //double t0 = time_time();
-                //fprintf(stderr, "unit=%%x N= %%i %%i %%i S = %%i %%i %%i %%i %%i %%i\\n", unit,
-                //Nz1, Nz0, Nx1,
-                //sy_0, sy_1,
-                //sx_0, sx_1,
-                //sz_0, sz_1
-                //);
                 switch(unit)
                 {
                     case 0x000: dgemm_(&N, &N, &Nz1, &Nz0, &Nx1, &a, y,
@@ -782,8 +756,6 @@ class GemmRelated(COp):
                                              "some matrix has no unit stride");
                              %(fail)s;
                 };
-                //fprintf(stderr, "Calling dgemm %%i %%i %%i %%i took %%f\\n",
-                //        unit, Nz1, Nz0, Nx1, time_time()- t0);
         """
 
     end_switch_typenum = """
@@ -822,7 +794,7 @@ class GemmRelated(COp):
         )
 
     def build_gemm_version(self):
-        return (13, blas_header_version())
+        return (14, blas_header_version())
 
 
 class Gemm(GemmRelated):
@@ -888,7 +860,7 @@ class Gemm(GemmRelated):
         return rval
 
     def make_node(self, *inputs):
-        inputs = list(map(ptb.as_tensor_variable, inputs))
+        inputs = list(map(as_tensor_variable, inputs))
 
         if any(not isinstance(i.type, DenseTensorType) for i in inputs):
             raise NotImplementedError("Only dense tensor types are supported")
@@ -976,8 +948,8 @@ class Gemm(GemmRelated):
         z_shape, _, x_shape, y_shape, _ = input_shapes
         return [
             (
-                pytensor.scalar.scalar_maximum(z_shape[0], x_shape[0]),
-                pytensor.scalar.scalar_maximum(z_shape[1], y_shape[1]),
+                pytensor.scalar.maximum(z_shape[0], x_shape[0]),
+                pytensor.scalar.maximum(z_shape[1], y_shape[1]),
             )
         ]
 
@@ -1063,7 +1035,7 @@ class Gemm(GemmRelated):
                 %(fail)s
             }
 
-            if(PyArray_MoveInto(x_new, %(_x)s) == -1)
+            if(PyArray_CopyInto(x_new, %(_x)s) == -1)
             {
                 %(fail)s
             }
@@ -1089,7 +1061,7 @@ class Gemm(GemmRelated):
                 %(fail)s
             }
 
-            if(PyArray_MoveInto(y_new, %(_y)s) == -1)
+            if(PyArray_CopyInto(y_new, %(_y)s) == -1)
             {
                 %(fail)s
             }
@@ -1135,7 +1107,7 @@ class Gemm(GemmRelated):
     def c_code_cache_version(self):
         gv = self.build_gemm_version()
         if gv:
-            return (7, *gv)
+            return (8, *gv)
         else:
             return gv
 
@@ -1148,322 +1120,6 @@ pprint.assign(gemm_inplace, FunctionPrinter(["gemm_inplace"]))
 pprint.assign(gemm_no_inplace, FunctionPrinter(["gemm_no_inplace"]))
 
 
-def res_is_a(fgraph, var, op, maxclients=None):
-    if maxclients is not None and var in fgraph.clients:
-        retval = len(fgraph.get_clients(var)) <= maxclients
-    else:
-        retval = True
-
-    return var.owner and var.owner.op == op and retval
-
-
-def _as_scalar(res, dtype=None):
-    """Return ``None`` or a `TensorVariable` of float type"""
-    if dtype is None:
-        dtype = config.floatX
-    if all(s == 1 for s in res.type.shape):
-        while res.owner and isinstance(res.owner.op, DimShuffle):
-            res = res.owner.inputs[0]
-        # may still have some number of True's
-        if res.type.ndim > 0:
-            rval = res.dimshuffle()
-        else:
-            rval = res
-        if rval.type.dtype in integer_dtypes:
-            # We check that the upcast of res and dtype won't change dtype.
-            # If dtype is float64, we will cast int64 to float64.
-            # This is valid when res is a scalar used as input to a dot22
-            # as the cast of the scalar can be done before or after the dot22
-            # and this will give the same result.
-            if pytensor.scalar.upcast(res.dtype, dtype) == dtype:
-                return ptb.cast(rval, dtype)
-            else:
-                return None
-
-        return rval
-
-
-def _is_real_matrix(res):
-    return (
-        res.type.dtype in ("float16", "float32", "float64")
-        and res.type.ndim == 2
-        and res.type.shape[0] != 1
-        and res.type.shape[1] != 1
-    )  # cope with tuple vs. list
-
-
-def _is_real_vector(res):
-    return (
-        res.type.dtype in ("float16", "float32", "float64")
-        and res.type.ndim == 1
-        and res.type.shape[0] != 1
-    )
-
-
-def _beta_L_plus_alpha_M(fgraph, beta, L, alpha, M, recurse_flip=True):
-    # print 'BETA L + ALPHA M', beta, L, alpha, M, recurse_flip
-    # EXPRESSION: (beta * L) + (alpha * M)
-
-    # we've already checked the client counts, now just make the type check.
-    # if res_is_a(M, _dot22, 1):
-    if M.owner and M.owner.op == _dot22:
-        Ml, Mr = M.owner.inputs
-        rval = [gemm_no_inplace(L, alpha, Ml, Mr, beta)]
-        return rval, M
-
-    # it also might be the case that there is a dimshuffle between the +
-    # and the dot22. local_dot_to_dot22 in particular will put in such things.
-    if (
-        M.owner
-        and isinstance(M.owner.op, DimShuffle)
-        and M.owner.inputs[0].owner
-        and isinstance(M.owner.inputs[0].owner.op, Dot22)
-    ):
-        MM = M.owner.inputs[0]
-        if M.owner.op.new_order == (0,):
-            # it is making a column MM into a vector
-            MMl, MMr = MM.owner.inputs
-            g = gemm_no_inplace(L.dimshuffle(0, "x"), alpha, MMl, MMr, beta)
-            rval = [g.dimshuffle(0)]
-            return rval, MM
-        if M.owner.op.new_order == (1,):
-            # it is making a row MM into a vector
-            MMl, MMr = MM.owner.inputs
-            g = gemm_no_inplace(L.dimshuffle("x", 0), alpha, MMl, MMr, beta)
-            rval = [g.dimshuffle(1)]
-            return rval, MM
-        if len(M.owner.op.new_order) == 0:
-            # it is making a row MM into a vector
-            MMl, MMr = MM.owner.inputs
-            g = gemm_no_inplace(L.dimshuffle("x", "x"), alpha, MMl, MMr, beta)
-            rval = [g.dimshuffle()]
-            return rval, MM
-
-    if recurse_flip:
-        return _beta_L_plus_alpha_M(fgraph, alpha, M, beta, L, recurse_flip=False)
-    else:
-        return False, False
-
-
-def _gemm_canonicalize(fgraph, r, scale, rval, maxclients):
-    # Tries to interpret node as a sum of scalars * (vectors or matrices)
-    def scaled(thing):
-        if scale == 1:
-            return thing
-        if scale == -1 and thing.type.dtype != "bool":
-            return -thing
-        else:
-            return scale * thing
-
-    if not isinstance(r.type, TensorType):
-        return None
-
-    if (r.type.ndim not in (1, 2)) or r.type.dtype not in (
-        "float16",
-        "float32",
-        "float64",
-        "complex64",
-        "complex128",
-    ):
-        rval.append(scaled(r))
-        return rval
-
-    if maxclients and len(fgraph.clients[r]) > maxclients:
-        rval.append((scale, r))
-        return rval
-
-    if r.owner and r.owner.op == sub:
-        _gemm_canonicalize(fgraph, r.owner.inputs[0], scale, rval, 1)
-        _gemm_canonicalize(fgraph, r.owner.inputs[1], -scale, rval, 1)
-
-    elif r.owner and r.owner.op == add:
-        for i in r.owner.inputs:
-            _gemm_canonicalize(fgraph, i, scale, rval, 1)
-
-    elif r.owner and r.owner.op == neg:
-        _gemm_canonicalize(fgraph, r.owner.inputs[0], -scale, rval, 1)
-
-    elif r.owner and r.owner.op == mul:
-        scalars = []
-        vectors = []
-        matrices = []
-        for i in r.owner.inputs:
-            if all(s == 1 for s in i.type.shape):
-                while i.owner and isinstance(i.owner.op, DimShuffle):
-                    i = i.owner.inputs[0]
-                if i.type.ndim > 0:
-                    scalars.append(i.dimshuffle())
-                else:
-                    scalars.append(i)
-            elif _is_real_vector(i):
-                vectors.append(i)
-            elif _is_real_matrix(i):
-                matrices.append(i)
-            else:
-                # just put the original arguments as in the base case
-                rval.append((scale, r))
-                return rval
-        if len(matrices) == 1:
-            assert len(vectors) == 0
-            m = matrices[0]
-            if len(scalars) == 0:
-                _gemm_canonicalize(fgraph, m, scale, rval, 1)
-            elif len(scalars) == 1:
-                _gemm_canonicalize(fgraph, m, scaled(scalars[0]), rval, 1)
-            else:
-                _gemm_canonicalize(
-                    fgraph, m, mul(scaled(scalars[0]), *scalars[1:]), rval, 1
-                )
-        elif len(vectors) == 1:
-            assert len(matrices) == 0
-            v = vectors[0]
-            if len(scalars) == 0:
-                _gemm_canonicalize(fgraph, v, scale, rval, 1)
-            elif len(scalars) == 1:
-                _gemm_canonicalize(fgraph, v, scaled(scalars[0]), rval, 1)
-            else:
-                _gemm_canonicalize(
-                    fgraph, v, mul(scaled(scalars[0]), *scalars[1:]), rval, 1
-                )
-        else:  # lets not open this up
-            rval.append((scale, r))
-    else:
-        rval.append((scale, r))
-    return rval
-
-
-def _factor_canonicalized(lst):
-    # remove duplicates from canonicalized list
-
-    # we only delete out of the right end of the list,
-    # once i has touched a list element, it is permantent
-    lst = list(lst)
-    # print 'FACTOR', lst
-    # for t in lst:
-    #    if not isinstance(t, (list, tuple)):
-    #        t = (t,)
-    #    for e in t:
-    #        try:
-    #            pytensor.printing.debugprint(e)
-    #        except TypeError:
-    #            print e, type(e)
-    i = 0
-    while i < len(lst) - 1:
-        try:
-            s_i, M_i = lst[i]
-        except Exception:
-            i += 1
-            continue
-
-        j = i + 1
-        while j < len(lst):
-            try:
-                s_j, M_j = lst[j]
-            except Exception:
-                j += 1
-                continue
-
-            if M_i is M_j:
-                s_i = s_i + s_j
-                lst[i] = (s_i, M_i)
-                del lst[j]
-            else:
-                j += 1
-        i += 1
-    return lst
-
-
-def _gemm_from_factored_list(fgraph, lst):
-    """
-    Returns None, or a list to replace node.outputs.
-
-    """
-    lst2 = []
-    # Remove the tuple that can't be cast correctly.
-    # This can happen when we try to cast a complex to a real
-    for sM in lst:
-        # Make every pair in list have matching dtypes
-        # sM can be a tuple of 2 elements or an PyTensor variable.
-        if isinstance(sM, tuple):
-            sm0, sm1 = sM
-            sm0 = ptb.as_tensor_variable(sm0)
-            if pytensor.scalar.upcast(sm0.dtype, sm1.dtype) == sm1.dtype:
-                lst2.append((ptb.cast(sm0, sm1.dtype), sM[1]))
-
-    lst = lst2
-
-    def item_to_var(t):
-        try:
-            s, M = t
-        except Exception:
-            return t
-        if s == 1:
-            return M
-        if s == -1:
-            return -M
-        return s * M
-
-    # Try every pair in the sM_list, trying to turn it into a gemm operation
-    for i in range(len(lst) - 1):
-        s_i, M_i = lst[i]
-
-        for j in range(i + 1, len(lst)):
-            s_j, M_j = lst[j]
-
-            if not M_j.type.in_same_class(M_i.type):
-                continue
-
-            # print 'TRYING', (s_i, M_i, s_j, M_j)
-
-            gemm_of_sM_list, old_dot22 = _beta_L_plus_alpha_M(
-                fgraph, s_i, M_i, s_j, M_j
-            )
-            # print 'GOT IT', gemm_of_sM_list
-            if gemm_of_sM_list:
-                assert len(gemm_of_sM_list) == 1
-                add_inputs = [
-                    item_to_var(input) for k, input in enumerate(lst) if k not in (i, j)
-                ]
-                add_inputs.extend(gemm_of_sM_list)
-                rval = [variadic_add(*add_inputs)]
-                return rval, old_dot22
-
-
-def _gemm_from_node2(fgraph, node):
-    """
-
-    TODO: In many expressions, there are many ways to turn it into a
-    gemm.  For example dot(a,b) + c + d.  This function should return all
-    of them, so that if one version of gemm causes a cycle in the graph, then
-    another application of gemm can be tried.
-
-    """
-    lst = []
-    t0 = time.perf_counter()
-    _gemm_canonicalize(fgraph, node.outputs[0], 1.0, lst, 0)
-    t1 = time.perf_counter()
-
-    if len(lst) > 1:
-        lst = _factor_canonicalized(lst)
-        t2 = time.perf_counter()
-        rval = _gemm_from_factored_list(fgraph, lst)
-        t3 = time.perf_counter()
-
-        # It can happen that _factor_canonicalized and
-        # _gemm_from_factored_list return a node with an incorrect
-        # type.  This happens in particular when one of the scalar
-        # factors forces the upcast of the whole expression.  In that
-        # case, we simply skip that candidate for Gemm.  This was
-        # discussed in
-        # http://groups.google.com/group/theano-dev/browse_thread/thread/a3096c82856e3ad5,
-        # but never made it into a trac ticket.
-
-        if rval and rval[0][0].type.in_same_class(node.outputs[0].type):
-            return rval, t1 - t0, t2 - t1, t3 - t2
-
-    return None, t1 - t0, 0, 0
-
-
 class Dot22(GemmRelated):
     """Compute a matrix-matrix product.
 
@@ -1474,8 +1130,8 @@ class Dot22(GemmRelated):
     check_input = False
 
     def make_node(self, x, y):
-        x = ptb.as_tensor_variable(x)
-        y = ptb.as_tensor_variable(y)
+        x = as_tensor_variable(x)
+        y = as_tensor_variable(y)
 
         if any(not isinstance(i.type, DenseTensorType) for i in (x, y)):
             raise NotImplementedError("Only dense tensor types are supported")
@@ -1490,16 +1146,8 @@ class Dot22(GemmRelated):
         outputs = [tensor(dtype=x.type.dtype, shape=(x.type.shape[0], y.type.shape[1]))]
         return Apply(self, [x, y], outputs)
 
-    def perform(self, node, inp, out):
-        x, y = inp
-        (z,) = out
-        try:
-            z[0] = np.asarray(np.dot(x, y))
-        except ValueError as e:
-            # The error raised by numpy has no shape information, we mean to
-            # add that
-            e.args = (*e.args, x.shape, y.shape)
-            raise
+    def perform(self, node, inputs, output_storage):
+        output_storage[0][0] = np.dot(*inputs)
 
     def infer_shape(self, fgraph, node, input_shapes):
         return [[input_shapes[0][0], input_shapes[1][1]]]
@@ -1667,8 +1315,8 @@ class BatchedDot(COp):
     gufunc_signature = "(b,m,k),(b,k,n)->(b,m,n)"
 
     def make_node(self, x, y):
-        x = ptb.as_tensor_variable(x)
-        y = ptb.as_tensor_variable(y)
+        x = as_tensor_variable(x)
+        y = as_tensor_variable(y)
 
         if not (
             isinstance(x.type, DenseTensorType) and isinstance(y.type, DenseTensorType)
@@ -1702,7 +1350,7 @@ class BatchedDot(COp):
 
         # Change dtype if needed
         dtype = pytensor.scalar.upcast(x.type.dtype, y.type.dtype)
-        x, y = ptb.cast(x, dtype), ptb.cast(y, dtype)
+        x, y = cast(x, dtype), cast(y, dtype)
         out = tensor(dtype=dtype, shape=out_shape)
         return Apply(self, [x, y], [out])
 
@@ -1887,7 +1535,7 @@ class BatchedDot(COp):
 
         return f"""
         int type_num = PyArray_DESCR({_x})->type_num;
-        int type_size = PyArray_DESCR({_x})->elsize; // in bytes
+        int type_size = PyArray_ITEMSIZE({_x}); // in bytes
 
         if (PyArray_NDIM({_x}) != 3) {{
             PyErr_Format(PyExc_NotImplementedError,
@@ -1947,14 +1595,14 @@ class BatchedDot(COp):
     def c_code_cache_version(self):
         from pytensor.tensor.blas_headers import blas_header_version
 
-        return (5, blas_header_version())
+        return (6, blas_header_version())
 
     def grad(self, inp, grads):
         x, y = inp
         (gz,) = grads
 
-        xgrad = batched_dot(gz, y.dimshuffle(0, 2, 1))
-        ygrad = batched_dot(x.dimshuffle(0, 2, 1), gz)
+        xgrad = _batched_dot(gz, y.dimshuffle(0, 2, 1))
+        ygrad = _batched_dot(x.dimshuffle(0, 2, 1), gz)
 
         # If x or y contain broadcastable dimensions but only one of
         # them know that a matching dimensions is broadcastable, the
@@ -1995,7 +1643,7 @@ class BatchedDot(COp):
                 iv1 = pytensor.graph.op.get_test_value(inputs[1])
             except TestValueError:
                 pytensor.graph.op.missing_test_message(
-                    "second input passed to BatchedDot.R_op has no test value"
+                    "second input passed to BatchedDot. R_op has no test value"
                 )
                 test_values_enabled = False
 
@@ -2004,8 +1652,7 @@ class BatchedDot(COp):
                     ev0 = pytensor.graph.op.get_test_value(eval_points[0])
                 except TestValueError:
                     pytensor.graph.op.missing_test_message(
-                        "first eval point passed to BatchedDot.R_op "
-                        "has no test value"
+                        "first eval point passed to BatchedDot. R_op has no test value"
                     )
                     test_values_enabled = False
             if eval_points[1]:
@@ -2013,8 +1660,7 @@ class BatchedDot(COp):
                     ev1 = pytensor.graph.op.get_test_value(eval_points[1])
                 except TestValueError:
                     pytensor.graph.op.missing_test_message(
-                        "second eval point passed to BatchedDot.R_op "
-                        "has no test value"
+                        "second eval point passed to BatchedDot.R_op has no test value"
                     )
                     test_values_enabled = False
 
@@ -2078,31 +1724,22 @@ def batched_dot(a, b):
             dot products in terms of batched matrix-matrix dot products, so
             it may be possible to further optimize for performance.
     """
-    a, b = ptb.as_tensor_variable(a), ptb.as_tensor_variable(b)
+    warnings.warn(
+        "batched_dot is deprecated. "
+        "Use `dot` in conjunction with `tensor.vectorize` or `graph.replace.vectorize_graph`",
+        FutureWarning,
+    )
+    a, b = as_tensor_variable(a), as_tensor_variable(b)
 
     if a.ndim == 0:
         raise TypeError("a must have at least one (batch) axis")
     elif b.ndim == 0:
         raise TypeError("b must have at least one (batch) axis")
-    elif a.ndim == 1:
-        return shape_padright(a, (b.ndim - 1)) * b
-    elif b.ndim == 1:
-        return a * shape_padright(b, (a.ndim - 1))
-    elif a.ndim > 3 or b.ndim > 3:
-        return batched_tensordot(a, b, [[a.ndim - 1], [np.maximum(1, b.ndim - 2)]])
-    else:
-        # If either a or b is a batched vector, expand dims and later squeeze them
-        expanded_axis = []
-        if a.ndim == 2:
-            a = expand_dims(a, axis=1)
-            expanded_axis.append(1)
-        if b.ndim == 2:
-            b = expand_dims(b, axis=2)
-            expanded_axis.append(2)
-        out = _batched_dot(a, b)
-        if expanded_axis:
-            out = out.squeeze(axis=expanded_axis)
-        return out
+
+    core_a = a[0].type()
+    core_b = b[0].type()
+    core_dot = dot(core_a, core_b)
+    return vectorize_graph(core_dot, replace={core_a: a, core_b: b})
 
 
 def batched_tensordot(x, y, axes=2):
@@ -2140,6 +1777,22 @@ def batched_tensordot(x, y, axes=2):
     reshapes to reduce the tensor dot product to a matrix or vector
     dot product.  Finally, it calls batched_dot to compute the result.
     """
-    from pytensor.tensor.math import _tensordot_as_dot
+    warnings.warn(
+        "batched_tensordot is deprecated. "
+        "Use `tensordot` in conjuction with `tensor.vectorize` or `graph.replace.vectorize_graph`",
+        FutureWarning,
+    )
 
-    return _tensordot_as_dot(x, y, axes, dot=batched_dot, batched=True)
+    if isinstance(axes, int):
+        core_axes = axes
+    else:
+        # Convert batched axes to core axes
+        core_axes_a = [a - 1 for a in normalize_axis_tuple(axes[0], x.type.ndim)]
+        core_axes = [a - 1 for a in normalize_axis_tuple(axes[1], y.type.ndim)]
+        core_axes = [core_axes_a, core_axes]
+
+    core_x = x[0].type()
+    core_y = y[0].type()
+    core_tensordot = tensordot(core_x, core_y, axes=core_axes)
+
+    return vectorize_graph(core_tensordot, replace={core_x: x, core_y: y})
